@@ -2,8 +2,7 @@
 # "It's not a mesh; it's a meshoid!" - Alfred H. Einstimes
 
 import numpy as np
-from scipy.spatial import cKDTree
-from numba import jit, vectorize, float32, float64, njit, guvectorize
+from scipy.spatial import KDTree
 from .grid_deposition import *
 from .kernel_density import *
 from .derivatives import *
@@ -14,7 +13,7 @@ class Meshoid:
 
     def __init__(
         self,
-        pos,
+        pos: np.ndarray,
         m=None,
         kernel_radius=None,
         des_ngb=None,
@@ -104,6 +103,10 @@ class Meshoid:
             self.vol = self.volnorm * self.kernel_radius**self.dim / self.des_ngb
             self.density = self.m / self.vol
 
+    def printv(self, *a, **k):
+        if self.verbose:
+            print(*a, **k)
+
     def reset_dweights(self):
         self.dweights = self.d2weights = self.dweights_3rdorder = None
 
@@ -118,8 +121,11 @@ class Meshoid:
         weighted: boolean, optional
             whether to kernel-weight the least-squares gradient solutions (default: True)
         """
-        if self.verbose:
-            print(f"Computing weights for derivatives of order {order}...")
+
+        self.printv(f"Computing weights for derivatives of order {order}...")
+
+        if self.ngb is None:
+            self.TreeUpdate()
 
         weights = gradient_weights(
             self.pos,
@@ -134,15 +140,11 @@ class Meshoid:
         if order == 1:
             self.dweights = weights
         elif order == 2:
-            self.d2weights, self.dweights_3rdorder = (
-                weights[:, self.dim :, :],
-                weights[:, : self.dim, :],
-            )
+            self.d2weights, self.dweights_3rdorder = (weights[:, self.dim :, :], weights[:, : self.dim, :])
 
     def BuildTree(self):
-        if self.verbose:
-            print("Building tree...")
-        self.tree = cKDTree(self.pos, boxsize=self.boxsize)
+        self.printv("Building tree...")
+        self.tree = KDTree(self.pos, boxsize=self.boxsize)
 
     def TreeUpdate(self):
         """
@@ -152,21 +154,18 @@ class Meshoid:
         if self.tree is None:
             self.BuildTree()
 
-        if self.verbose:
-            print("Finding neighbors...")
+        self.printv("Finding neighbors...")
         self.ngbdist, self.ngb = self.tree.query(
-            self.pos[self.particle_mask], round(self.des_ngb * 1.1), workers=self.n_jobs
+            self.pos[self.particle_mask], round(self.des_ngb * 1.5), workers=self.n_jobs
         )
 
-        if self.verbose:
-            print("Neighbours found!")
+        self.printv("Neighbours found!")
 
-        if self.verbose:
-            print("Iterating for smoothing lengths...")
+        self.printv("Iterating for smoothing lengths...")
 
         self.kernel_radius = HsmlIter(self.ngbdist, des_ngb=self.des_ngb, error_norm=1e-13, dim=self.dim)
-        if self.verbose:
-            print("Smoothing lengths found!")
+
+        self.printv("Smoothing lengths found!")
 
         self.density = self.des_ngb * self.m[self.particle_mask] / (self.volnorm * self.kernel_radius**self.dim)
         self.vol = self.m[self.particle_mask] / self.density
@@ -310,11 +309,7 @@ class Meshoid:
         shape (N,3) array containing the curl of vector field v
         """
         dv = self.D(v[self.particle_mask])
-        return np.c_[
-            dv[:, 2, 1] - dv[:, 1, 2],
-            dv[:, 0, 2] - dv[:, 2, 0],
-            dv[:, 1, 0] - dv[:, 0, 1],
-        ]
+        return np.c_[dv[:, 2, 1] - dv[:, 1, 2], dv[:, 0, 2] - dv[:, 2, 0], dv[:, 1, 0] - dv[:, 0, 1]]
 
     def Div(self, v):
         """
@@ -410,9 +405,16 @@ class Meshoid:
             self.BuildTree()
         if order < 0 or order > 2:
             raise NotImplementedError("Reconstruction of order > 2 not supported.")
+
+        if order == 1:
+            if self.dweights is None:
+                self.ComputeDWeights(1)
+        elif order == 2:
+            if self.d2weights is None:
+                self.ComputeDWeights(2)
         # get nearest neighbor of each target point
-        target_neighbors = self.tree.query(target_points, workers=self.n_jobs)[1]
-        unique_neighbors, neighbor_idx = np.unique(target_neighbors, return_inverse=True)
+        _, target_neighbors = self.tree.query(target_points, workers=self.n_jobs)
+        #        unique_neighbors, neighbor_idx = np.unique(target_neighbors, return_inverse=True)
 
         # get value of f at each nearest neighbor
         f_target = np.take(f, target_neighbors, axis=0)
@@ -420,6 +422,7 @@ class Meshoid:
             return f_target
 
         # 1st-order reconstruction
+        self.printv(target_points, self.pos.take(target_neighbors, axis=0))
         dx = target_points - self.pos.take(target_neighbors, axis=0)
         if self.boxsize is not None:
             dx = nearest_image_v(dx, self.boxsize)
@@ -427,13 +430,13 @@ class Meshoid:
         # self.particle_mask = unique_neighbors
         # self.TreeUpdate()  # update neighbor lists to just the target neighbors
         # self.reset_dweights()  # reset derivative weights if already computed
-        gradf_neighbors = self.D(f, order=order).take(neighbor_idx, axis=0)
+        gradf_neighbors = self.D(f, order=order).take(target_neighbors, axis=0)
         f_target += np.einsum("ij,i...j->i...", dx, gradf_neighbors)
         if order == 1:
             return f_target
 
         # 2nd order reconstruction
-        d2f_neighbors = self.D2(f).take(neighbor_idx, axis=0)
+        d2f_neighbors = self.D2(f).take(target_neighbors, axis=0)
         f_target += 0.5 * np.einsum("ij,ij...->i...", dx * dx, d2f_neighbors[..., :3])  # pure 2nd derivative terms
         for dim in range(self.dim):
             f_target += np.einsum(
@@ -443,16 +446,7 @@ class Meshoid:
             )  # mixed terms
         return f_target
 
-    def Slice(
-        self,
-        f,
-        size=None,
-        plane="z",
-        center=None,
-        res: int = 128,
-        order: int = 0,
-        return_grid: bool = False,
-    ):
+    def Slice(self, f, size=None, plane="z", center=None, res: int = 128, order: int = 0, return_grid: bool = False):
         """
         Gives the value of a function f deposited on a 2D Cartesian grid slicing
         through the 3D domain.
@@ -505,15 +499,7 @@ class Meshoid:
             return x, y, f_grid.reshape(shape)
         return f_grid.reshape(shape)
 
-    def InterpToGrid(
-        self,
-        f,
-        size=None,
-        center=None,
-        res: int = 128,
-        order: int = 1,
-        return_grid: bool = False,
-    ):
+    def InterpToGrid(self, f, size=None, center=None, res: int = 128, order: int = 1, return_grid: bool = False):
         """
         Interpolates the quantity f defined on the meshoid to the cell centers
         of a 3D Cartesian grid
@@ -596,32 +582,21 @@ class Meshoid:
 
         return f_grid
 
-    def SurfaceDensity(
-        self,
-        f=None,
-        size=None,
-        # plane="z",
-        center=None,
-        res=128,
-        smooth_fac=1.0,
-        conservative=False,
-    ):
+    def SurfaceDensity(self, f=None, size=None, center=None, res=128, smooth_fac=1.0, conservative=False):
         """
         Computes the surface density of a quantity f defined on the meshoid on a grid of sightlines. e.g. if f is the particle masses, you will get mass surface density.
 
         Parameters
         ----------
-        f :
+        f: array_like, optional
             the quantity you want the surface density of (default: particle mass)
-        size :
+        size: float, optional
             the side length of the window of sightlines (default: None, will use the meshoid's predefined side length')
-        plane :
-            the direction you want to project along, of x, y, or z (default: 'z')
-        center :
+        center: array_like, optional
             (2,) or (3,) array containing the coordaintes of the center of the grid (default: None, will use the meshoid's predefined center)
-        res :
+        res: int, optional
             the resolution of the grid of sightlines (default: 128)
-        smooth_fac :
+        smooth_fac: float, optional
             smoothing lengths are increased by this factor (default: 1.)
 
         Returns
@@ -646,23 +621,64 @@ class Meshoid:
             conservative=conservative,
         )
 
+    def MassWeightedAverage(self, f, masses=None, size=None, center=None, res=128, smooth_fac=1.0, conservative=False):
+        """
+        Computes the surface density of a quantity f defined on the meshoid on a grid of sightlines. e.g. if f is the particle masses, you will get mass surface density.
+
+        Parameters
+        ----------
+        f: array_like
+            The quantity you want the mass-weighted average of
+        masses: array_like, optional
+            Shape (N,) array of masses, or in general arbitrary weights
+        size: float, optional
+            the side length of the window of sightlines (default: None, will use the meshoid's predefined side length')
+        center: array_like, optional
+            (2,) or (3,) array containing the coordaintes of the center of the grid (default: None, will use the meshoid's predefined center)
+        res: int, optional
+            the resolution of the grid of sightlines (default: 128)
+        smooth_fac: float, optional
+            smoothing lengths are increased by this factor (default: 1.)
+
+        Returns
+        -------
+        (res,res) array containing the column densities integrated along sightlines
+        """
+        if f is None:
+            f = self.m
+        if center is None:
+            center = self.center
+        if size is None:
+            size = self.L
+
+        args = (
+            self.pos,
+            np.clip(smooth_fac * self.kernel_radius, 2 * size / res, 1e100),
+            center,
+            size,
+            res,
+            self.boxsize,
+            (False if self.n_jobs == 1 else True),
+            conservative,
+        )
+
+        return GridSurfaceDensity(f * masses, *args) / GridSurfaceDensity(masses, *args)
+
     def ProjectedAverage(self, f, size=None, center=None, res=128, smooth_fac=1.0):  # plane="z",
         """
         Computes the average value of a quantity f along a Cartesian grid of sightlines from +/- infinity.
 
         Parameters
         ----------
-        f :
+        f: array_like, optional
             (N,) array containing the quantity you want the average of
-        size :
+        size: float, optional
             the side length of the window of sightlines (default: None, will use the meshoid's predefined side length')
-        plane :
-            the direction you want to project along, of x, y, or z (default: 'z')
-        center :
+        center: array_like, optional
             (2,) or (3,) array containing the coordaintes of the center of the grid (default: None, will use the meshoid's predefined center)
-        res :
+        res: int, optional
             the resolution of the grid of sightlines (default: 128)
-        smooth_fac :
+        smooth_fac: float, optional
             smoothing lengths are increased by this factor (default: 1.)
 
         Returns
@@ -683,7 +699,7 @@ class Meshoid:
             self.boxsize,
         )
 
-    def Projection(self, f, size=None, center=None, res=128, smooth_fac=1.0):  # plane="z",
+    def Projection(self, f, size=None, center=None, res=128, smooth_fac=1.0):
         """
         Computes the integral of quantity f along a Cartesian grid of sightlines
         from +/- infinity. e.g. plugging in 3D density for f will return
@@ -691,18 +707,15 @@ class Meshoid:
 
         Parameters
         ----------
-        f :
-            (N,) array containing the quantity you want the projected integral of
-        size :
+        f: array_like, optional
+            (N,) array containing the quantity you want the average of
+        size: float, optional
             the side length of the window of sightlines (default: None, will use the meshoid's predefined side length')
-        plane :
-            the direction you want to project along, of x, y, or z (default: 'z')
-        center :
-            (2,) or (3,) array containing the coordaintes of the center of the
-            grid (default: None, will use the meshoid's predefined center)
-        res :
+        center: array_like, optional
+            (2,) or (3,) array containing the coordaintes of the center of the grid (default: None, will use the meshoid's predefined center)
+        res: int, optional
             the resolution of the grid of sightlines (default: 128)
-        smooth_fac :
+        smooth_fac: float, optional
             smoothing lengths are increased by this factor (default: 1.)
 
         Returns
@@ -713,7 +726,7 @@ class Meshoid:
             center = self.center
         if size is None:
             size = self.L
-        return GridAverage(
+        return GridSurfaceDensity(
             f * self.vol,
             self.pos,
             np.clip(smooth_fac * self.kernel_radius, 2 * size / res, 1e100),
@@ -745,7 +758,7 @@ class Meshoid:
             bandwidth = self.SmoothingLength()
 
         f = np.zeros_like(grid)
-        gtree = cKDTree(np.c_[grid,])
+        gtree = KDTree(np.c_[grid,])
         for d, bw in zip(self.pos, bandwidth):
             ngb = gtree.query_ball_point(d, bw)
             ngbdist = np.abs(grid[ngb] - d)
